"""
Loads the election buddy schema
"""

from rcvformats.common.electionbuddyparser import ElectionBuddyData
from rcvformats.schemas.base import Schema


class SchemaV0(Schema):
    """
    ElectionBuddy's CSV format
    """

    def version(self):
        return 'unversioned:0'

    def _validate_file_object(self, file_object):
        try:
            ElectionBuddyData(file_object)
            return True
        except Exception as exception:  # pylint:disable=broad-except
            self._last_error = exception
            return False



"""
Integration tests for conversions between file formats
"""

import os
import json
import nose

from rcvformats.conversions import automatic
from rcvformats.conversions import dominion
from rcvformats.conversions import electionbuddy
from rcvformats.conversions import opavote
from rcvformats.conversions.ut_without_transfers import UTWithoutTransfersConverter


def _assert_conversion_correct(file_in, file_out, converter):
    """ Asserts that converter.convert_to_ut(file_in) = file_out """
    nose.tools.assert_dict_equal.__self__.maxDiff = None
    actual_data = converter.convert_to_ut_and_validate(file_in)
    with open(file_out, 'r', encoding='utf-8') as file_obj:
        expected_data = json.load(file_obj)
    nose.tools.assert_dict_equal(actual_data, expected_data)


def _files_in_dir(input_dir):
    """ Yields all CSV and json files in dir """
    for filename in os.listdir(input_dir):
        if not filename.endswith('.json') and \
                not filename.endswith('.csv') and \
                not filename.endswith('.xlsx'):
            continue
        yield filename


def _assert_auto_gives_same_result_as(input_dir, direct_converter):
    """
    Asserts that the AutomaticConverter creates the same results as the provided one
    for every file in the provided directory, plus a transfer dict
    """
    auto_converter = automatic.AutomaticConverter()
    add_transfer_converter = UTWithoutTransfersConverter(allow_guessing=False)

    for filename in _files_in_dir(input_dir):
        filepath = os.path.join(input_dir, filename)
        auto_data = auto_converter.convert_to_ut_and_validate(filepath)
        expected_data_without_xfer = direct_converter.convert_to_ut_and_validate(filepath)
        expected_data = add_transfer_converter.fill_in_tally_data(expected_data_without_xfer)

        nose.tools.assert_dict_equal(auto_data, expected_data)


def test_electionbuddy_conversions_succeed():
    """ Converts electionbuddy CSV to standard format """
    filenames = [
        'testdata/inputs/electionbuddy/standard.csv',
        'testdata/inputs/electionbuddy/multiwinner.csv',
        'testdata/inputs/electionbuddy/without-abstentions.csv'
    ]
    converter = electionbuddy.ElectionBuddyConverter()
    for filename in filenames:
        converter.convert_to_ut_and_validate(filename)


def test_opavote_conversion_accurate():
    """ Converts opavote JSON to standard format """
    file_in = 'testdata/inputs/opavote/fairvote.json'
    file_out = 'testdata/conversions/from-opavote.json'
    converter = opavote.OpavoteConverter()
    _assert_conversion_correct(file_in, file_out, converter)


def test_electionbuddy_conversion_accurate():
    """ Converts electionbuddy CSV to the standard format """
    file_in = 'testdata/inputs/electionbuddy/standard.csv'
    file_out = 'testdata/conversions/from-electionbuddy.json'
    converter = electionbuddy.ElectionBuddyConverter()
    _assert_conversion_correct(file_in, file_out, converter)

    file_in = 'testdata/inputs/electionbuddy/standard-with-threshold.csv'
    file_out = 'testdata/conversions/from-electionbuddy-with-threshold.json'
    converter = electionbuddy.ElectionBuddyConverter()
    _assert_conversion_correct(file_in, file_out, converter)


def test_dominion_conversion_accurate():
    """ Converts dominion XLSX to the standard format """
    file_in = 'testdata/inputs/dominion/las-cruces-mayor.xlsx'
    file_out = 'testdata/conversions/from-dominion.json'
    converter = dominion.DominionConverter()
    _assert_conversion_correct(file_in, file_out, converter)


def test_automatic_conversions_universal_tabulator():
    """ Tests that the automatic conversion works when given Universal Tabulator data """
    converter = automatic.AutomaticConverter()
    input_dir = 'testdata/inputs/universal-tabulator'

    for filename in _files_in_dir(input_dir):
        filepath = os.path.join(input_dir, filename)
        output_data = converter.convert_to_ut_and_validate(filepath)

        # Must be unchanged
        with open(filepath, 'r', encoding='utf-8') as input_file:
            input_data = json.load(input_file)

        nose.tools.assert_dict_equal(input_data, output_data)


def test_automatic_conversions_opavote():
    """ Tests that the automatic conversion works when given Opavote data """
    converter = opavote.OpavoteConverter()
    input_dir = 'testdata/inputs/opavote'
    _assert_auto_gives_same_result_as(input_dir, converter)


def test_automatic_conversions_dominion():
    """ Tests that the automatic conversion works when given Dominion data """
    converter = dominion.DominionConverter()
    input_dir = 'testdata/inputs/dominion'
    _assert_auto_gives_same_result_as(input_dir, converter)


def test_automatic_conversions_electionbuddy():
    """ Tests that the automatic conversion works when given ElectionBuddy data """
    converter = electionbuddy.ElectionBuddyConverter()
    input_dir = 'testdata/inputs/electionbuddy'
    _assert_auto_gives_same_result_as(input_dir, converter)


def _does_all_single_elim_have_transfer_data(data):
    for result in data['results']:
        tally_results = result['tallyResults']
        if len(tally_results) != 1:
            continue

        tally_result = tally_results[0]
        if 'elected' in tally_result:
            continue

        if not tally_result['transfers']:
            return False
    return True


def _does_all_batch_elim_have_transfer_data(data):
    for result in data['results']:
        tally_results = result['tallyResults']
        if len(tally_results) == 1:
            continue

        for tally_result in tally_results:
            if 'elected' in tally_result:
                continue
            if not tally_result['transfers']:
                return False
    return True


def test_add_xfer_without_fake_data():
    """ Tests allow_guessing=False in UT Without Transfers """
    converter = UTWithoutTransfersConverter(allow_guessing=False)
    input_filename = 'testdata/inputs/ut-without-transfers/nyc-batch-elim.json'
    with_transfers = converter.convert_to_ut(input_filename)
    assert _does_all_single_elim_have_transfer_data(with_transfers)
    assert not _does_all_batch_elim_have_transfer_data(with_transfers)


def test_add_xfer_with_fake_data():
    """ Tests allow_guessing=True in UT Without Transfers """
    converter = UTWithoutTransfersConverter(allow_guessing=True)
    input_filename = 'testdata/inputs/ut-without-transfers/nyc-batch-elim.json'
    with_transfers = converter.convert_to_ut(input_filename)
    assert _does_all_single_elim_have_transfer_data(with_transfers)
    assert _does_all_batch_elim_have_transfer_data(with_transfers)



"""
Reads an Dominion XLSX results file, writes to the standard format
"""
from openpyxl import load_workbook
from openpyxl.cell.cell import MergedCell
from openpyxl.cell.read_only import EmptyCell, ReadOnlyCell

from rcvformats.conversions.base import GenericGuessAtTransferConverter


class DominionConverter(GenericGuessAtTransferConverter):
    """
    Parses the dominion file format as exemplified in /testdata/inputs/dominion
    These are .xlsx files
    """

    # Define constants
    DATE_CELL = 'A9'

    class RoundInfo:  # pylint: disable=too-few-public-methods
        """
        Data for parsing a round:
        Because columns are hapharzadly merged, there is no straightforward mapping from
        round number to the corresponding column. This little struct keeps track of it for us.
        """
        round_num = None
        column = None

    class RowConstants():
        """
        Data for the row number that various items are on
        """
        # Define constants
        # the -12 is because New Mexico files have 12 rows of headers,
        # so the first number is the actual row number, then we subtract num_header_rows,
        # just for readability.
        SEAT_TITLE_NUM_ROWS_AFTER_HEADER = 12 - 12
        ROUND_LABELS_NUM_ROWS_AFTER_HEADER = 32 - 12
        FIRST_CANDIDATE_NUM_ROWS_AFTER_HEADER = 34 - 12
        ROW_AFTER_INACTIVE_FOR_THRESHOLD = 1

        def __init__(self):
            # The row that includes the seat title in column 1
            self.seat_title = None

            # The row that includes the round labels in the primary table
            self.round_label = None

            # The row that includes the candidate names in column 1
            self.first_candidate = None

            # Row that has the threshold at each round
            self.threshold = None

            # Row that has the number of inactive ("non transferrable") ballots at each round
            self.inactive_ballots = None

        def find_rows_before_summary_table(self, sheet):
            """
            Fills in values for rows before the summary table
            """
            offset = self._count_num_header_rows(sheet)
            self.round_label = self.ROUND_LABELS_NUM_ROWS_AFTER_HEADER + offset
            self.first_candidate = self.FIRST_CANDIDATE_NUM_ROWS_AFTER_HEADER + offset
            self.seat_title = self.SEAT_TITLE_NUM_ROWS_AFTER_HEADER + offset

        def find_rows_after_summary_table(self, sheet, num_candidates):
            """
            Fills in values for rows after the summary table,
            which requires needing to know the number of candidates
            """
            self.inactive_ballots = self._find_row_of_inactive_ballots(sheet, num_candidates)
            self.threshold = self._find_row_of_threshold(sheet, self.inactive_ballots)

        @classmethod
        def _count_num_header_rows(cls, sheet):
            """
            The number of header rows can vary.
            Looks for the first left-aligned row, which is row 12 or 13 in all data
            we've seen.
            """
            min_header_rows = 6
            max_header_rows = 50
            for row in range(min_header_rows, max_header_rows):
                # Empty rows are merged rows - ignore them
                if sheet.cell(row, 1).value is None:
                    continue

                # Center-aligned rows are header rows - keep going
                if sheet.cell(row, 1).alignment.horizontal == 'center':
                    continue

                return row
            raise Exception("Could not find the end of the headers")

        def _find_row_of_inactive_ballots(self, sheet, num_candidates):
            """
            Returns row number labeled "Non-Transferrable Total"
            """
            row = self.first_candidate
            row += num_candidates

            min_num_rows = 1
            max_num_rows = 6
            for row in range(row + min_num_rows, row + max_num_rows):
                if sheet.cell(row, 1).value == "Non Transferable Total":
                    return row
            raise Exception("Could not find the end of the non-transferable rows")

        def _find_row_of_threshold(self, sheet, inactive_row):
            """
            Returns row number labeled "Threshold"
            """
            row = inactive_row + self.ROW_AFTER_INACTIVE_FOR_THRESHOLD
            assert sheet.cell(row, 1).value == "Threshold"
            return row

    def __init__(self):
        super().__init__()

        # These fields will be filled out as data is loaded

        # The loaded spreadsheet
        self.sheet = None

        # Row constants
        self.row_constants = self.RowConstants()

        # names of each candidate
        self.candidates = None

        # list of RoundInfo, one per round
        self.data_per_round = None

    def _convert_file_object_to_ut(self, file_object):
        workbook = load_workbook(file_object)  # note: somehow, readonly is 2x slower
        self.sheet = workbook.active

        self.row_constants.find_rows_before_summary_table(self.sheet)
        self.candidates = self._parse_candidates()
        self.row_constants.find_rows_after_summary_table(self.sheet, len(self.candidates))

        self.data_per_round = self._parse_rounds()

        config = self._parse_config()
        results = self._get_vote_counts_per_candidate()
        urcvt_data = {'config': config, 'results': results}

        workbook.close()

        self._postprocess_remove_last_round_elimination(urcvt_data)
        self._postprocess_set_threshold(urcvt_data)
        return urcvt_data

    def _parse_config(self):
        """
        Returns the URCV config format
        """
        date_with_time = self.sheet[self.DATE_CELL].value
        date = str(date_with_time.date())
        seat = self.sheet.cell(self.row_constants.seat_title, 1).value

        config = {
            'contest': seat,
            'date': date,
            'office': seat
        }
        return config

    def _parse_rounds(self):
        """
        Rounds are curious - they are separated by a varying number of columns,
        usually 3 except for the first few, where the initial columns have been
        merged seemingly randomly (but in actually: twice in round 1, once in round 2)

        Therefore, this returns a pair of (Round, Column #) to get the column for
        the number of votes in each Round
        """
        col = 3
        data_per_round = []
        max_cols = 1500
        while True:
            cell = self.sheet.cell(self.row_constants.round_label, col)
            value = cell.value

            col += 1
            if col >= max_cols:
                raise Exception("This document is not in the correct format..."
                                "or there are more than 500 rounds")

            # Is this a merged cell? If so, ignore it.
            if isinstance(cell, MergedCell):
                continue

            # Have we reached the end of the table?
            if isinstance(cell, EmptyCell):
                break

            if value is None:
                # If we open this as read-only, then None represents a
                # merged cell and we should keep reading
                if isinstance(cell, ReadOnlyCell):
                    continue

                # Otherwise, it represents the end of the table
                break

            num_rounds = len(data_per_round)

            # Are we still in the same round?
            if value == 'Round ' + str(num_rounds):
                continue

            # If not, we must be in the next round. Increment and carry on.
            assert value == 'Round ' + str(num_rounds + 1)
            roundinfo = self.RoundInfo()
            roundinfo.round_num = num_rounds
            roundinfo.column = col - 1
            data_per_round.append(roundinfo)

        assert len(data_per_round) >= 1
        return data_per_round

    def _parse_candidates(self):
        """
        Grabs the list of candidate names from the summary table
        """
        end_of_candidates_marker = 'Continuing Ballots Total'
        max_num_rows = 500
        candidate_names = []
        row = self.row_constants.first_candidate
        while True:
            candidate_name = self.sheet['A' + str(row)].value
            if candidate_name == end_of_candidates_marker:
                break
            if row == max_num_rows:
                raise Exception("This document is not in the correct format..."
                                "or there are more than 500 candidates")
            cell = self.sheet['A' + str(row)]
            name = cell.value
            candidate_names.append(name)
            row += 1
        assert len(candidate_names) >= 1
        return candidate_names

    @classmethod
    def _is_eliminated_color(cls, color):
        """
        Is this cell colored red for elimination?
        Note: only call this if you've checked that the cell is filled solid.
        It may be the case that the cell is red, but not filled, in which case
        it appears white and is not eliminated.
        """
        return color == 'FFFFABAB'

    @classmethod
    def _is_elected_color(cls, color):
        """
        Is this cell colored greed, noting it is elected?
        Note: only call this if you've checked that the cell is filled solid.
        It may be the case that the cell is red, but not filled, in which case
        it appears white and is not eliminated.
        """
        return color == 'FF89CC89'

    def _parse_tally_for_round_at_column(self, col, eliminated_names):
        """ Creates a 'tally' and 'tallyResults' struct for the given round """
        starting_candidate_row = self.row_constants.first_candidate
        tally = {}
        tally_results = []
        for i, name in enumerate(self.candidates):
            if name in eliminated_names:
                continue

            row = starting_candidate_row + i
            cell = self.sheet.cell(row, col)
            vote_count = cell.value
            tally[name] = vote_count

            fill_type = cell.fill.fill_type
            if fill_type is not None:
                bg_color = cell.fill.bgColor.rgb
                if self._is_eliminated_color(bg_color):
                    eliminated_names.add(name)
                    tally_results.append({'eliminated': name})
                elif self._is_elected_color(bg_color):
                    tally_results.append({'elected': name})
                else:
                    # Just make sure we have no rogue colors
                    assert bg_color == '00000000'

        # Add inactive ballots
        tally['Inactive Ballots'] = self.sheet.cell(self.row_constants.inactive_ballots, col).value

        return tally, tally_results, eliminated_names

    def _get_vote_counts_per_candidate(self):
        results = []
        eliminated_names = set()
        for round_info in self.data_per_round:
            col = round_info.column
            tally, tally_results, eliminated_names = \
                self._parse_tally_for_round_at_column(col, eliminated_names)
            results.append({
                'round': round_info.round_num + 1,
                'tally': tally,
                'tallyResults': tally_results})
        return results

    @classmethod
    def _postprocess_remove_last_round_elimination(cls, data):
        """
        When there are two candidates left, Dominion marks the loser among them as
        "eliminated", whereas the URCVT format does not.
        Updates data to remove any last-round eliminations
        """
        last_round_tally_results = data['results'][-1]['tallyResults']
        last_round_tally_results = [t for t in last_round_tally_results if 'eliminated' not in t]
        data['results'][-1]['tallyResults'] = last_round_tally_results

    def _postprocess_set_threshold(self, data):
        """
        Set the threshold based on (last round active votes) / (num winners + 1)
        Which is also just listed on the table of per-round info
        """
        last_round_col = self.data_per_round[-1].column
        threshold = self.sheet.cell(self.row_constants.threshold, last_round_col).value
        data['config']['threshold'] = threshold



"""
Attempts to convert any file to the standard format.
If you know the file format, you should not use this - it will
loop through all schemas which will be needlessly slow.
"""

import json

from rcvformats.conversions.base import CouldNotConvertException
from rcvformats.conversions.dominion import DominionConverter
from rcvformats.conversions.electionbuddy import ElectionBuddyConverter
from rcvformats.conversions.opavote import OpavoteConverter
from rcvformats.conversions.ut_without_transfers import UTWithoutTransfersConverter
from rcvformats.conversions.base import Converter


class AutomaticConverter(Converter):
    """ Interface for converters """

    def __init__(self):
        # In order of likelihood of a hit - just my guess
        self.converters = [
            DominionConverter,
            ElectionBuddyConverter,
            OpavoteConverter
        ]

        super().__init__()

    def _convert_file_object_to_ut(self, file_object):
        # If it matches the schema already, load and return the data
        if self.ut_schema.validate(file_object):
            file_object.seek(0)
            return json.load(file_object)

        # Otherwise, try each converter - skipping schemas for speed
        for converter in self.converters:
            file_object.seek(0)
            try:
                data = converter().convert_to_ut(file_object)
                converter = UTWithoutTransfersConverter(allow_guessing=False)
                return converter.fill_in_tally_data(data)
            except CouldNotConvertException:
                continue

        # If it failed, accumulate all errors from schemas
        error_message = "When trying to parse as the Universal Tabulator schema, " +\
                        f"received error: \"{str(self.ut_schema.last_error())}\". " +\
            "Further, it did not match any other known format."
        raise CouldNotConvertException(error_message)



"""
Loads the universal tabulator schema
"""

from rcvformats.schemas.base import GenericJsonSchema, DataError


class SchemaV0(GenericJsonSchema):
    """ Schema for the initial version of the Universal RCV Tabulator """

    @property
    def schema_filename(self):
        return 'universaltabulator.schema.json'

    def version(self):
        return "Unversioned:V0"

    def ensure_data_is_logical(self, data):
        """
        Various checks to ensure the data is sane - though it cannot catch everything,
        we have tried to place the most common errors here
        """
        self.check_last_round_eliminations(data)
        self.check_candidate_leaves_after_elimination(data)

    @classmethod
    def check_last_round_eliminations(cls, data):
        """
        No eliminations allowed on the last round
        """
        last_round_tally_results = data['results'][-1]['tallyResults']
        for tally_result in last_round_tally_results:
            if 'eliminated' in tally_result:
                raise DataError("There cannot be an elimination on the last round. "
                                "All eliminations require one additonal round to signify where the "
                                "votes have been transferred to.")

    @classmethod
    def check_candidate_leaves_after_elimination(cls, data):
        """
        After a candidate is eliminated, ensure they are not listed later as having
        zero votes. They should be removed from the list.
        """
        eliminated_so_far = set()
        for round_num, result in enumerate(data['results']):
            tally = result['tally']
            for name in tally:
                if name in eliminated_so_far:
                    raise DataError(
                        f"Found {name} in Round {round_num+1}, though they were "
                        "already eliminated. After a candidate is eliminated, they should "
                        "be removed from all future vote tallies.")

            tally_results = result['tallyResults']
            for tally_result in tally_results:
                if 'eliminated' in tally_result:
                    eliminated_so_far.add(tally_result['eliminated'])



"""
Safety tests: enforce security measures here
"""
import nose
from openpyxl import DEFUSEDXML


def test_electionbuddy_conversions_succeed():
    """ Ensures defusedXml is installed to prevent malicious XML attacks """
    nose.tools.assert_true(DEFUSEDXML)



"""
Helper class for loading Electionbuddy CSVs
"""


class ElectionBuddyData:
    """
    Structure to hold the raw data read from the file.
    Parses the data directly into a simple format.
    Used for both conversion and schema validation.
    """

    def __init__(self, file_obj):
        self.file_obj = file_obj
        self.title = self.read_line_as_str()
        self.read_line_as_str()  # line break

        self.position = self.read_line_as_str()
        self.read_line_as_str()  # stars
        self.read_line_as_str()  # line break

        self.read_each_round()

    def read_line_as_str(self):
        """ Read line, and if it's bytes, decode to utf-8 """
        line = self.file_obj.readline()
        if isinstance(line, bytes):
            return line.decode("utf-8")
        return line

    def peek_line(self):
        """ Peeks at the next line without advancing """
        pos = self.file_obj.tell()
        line = self.read_line_as_str()
        self.file_obj.seek(pos)
        return line

    def read_each_round(self):
        """ Start iterating over the CSV for each round """
        self.rounds = []
        while self.file_obj:
            round_text = self.read_line_as_str()
            if not round_text.startswith("Round"):
                assert len(self.rounds) > 0
                break

            self.rounds.append(self.read_round())

    def read_round(self):
        """ Reads the CSV data for the next round """
        headers = self.read_line_as_str()
        assert headers.strip() == "Candidate,Votes,Percentage"

        candidates = {}
        threshold = None
        while True:
            line = self.read_line_as_str()
            if line.strip() == "":
                break
            candidate, votes, _ = line.split(',')
            candidates[candidate] = float(votes)

        # Eat summary lines (votes tallied, abstentions, newline)
        line = self.read_line_as_str()
        assert line.startswith('Votes tallied')

        # There are two optional lines: abstentions and threshold.
        # Check for each.
        for _ in range(2):
            line = self.peek_line()
            if line.startswith('Abstentions: '):
                # Abstentions line is optional
                self.read_line_as_str()
            elif line.startswith('Threshold: '):
                # Threshold line is optional
                line = self.read_line_as_str()
                threshold_str = line[len('Threshold: '):]
                threshold = float(threshold_str)
            else:
                assert line.strip() == ''
                break

        # Check for newline
        line = self.read_line_as_str()
        assert line.strip() == ''

        return {'candidates': candidates,
                'threshold': threshold}



"""
Reads an ElectionBuddy CSV results file, writes to the standard format
"""

from rcvformats.common.electionbuddyparser import ElectionBuddyData
from rcvformats.conversions.base import GenericGuessAtTransferConverter


class ElectionBuddyConverter(GenericGuessAtTransferConverter):
    """
    Reads an electionbuddy-formatted CSV file. Note that this
    use a generic text file reader, not a CSV reader, because
    it's not really a CSV file - it has parts of it that are,
    but it also has miscellaneous title lines.
    """

    def _convert_file_object_to_ut(self, file_object):
        raw_data = ElectionBuddyData(file_object)

        # Create configuration, assuming date of election is the file creation date
        config = {
            'contest': raw_data.title.strip(),
            'threshold': str(self._get_threshold(raw_data))
        }

        # Loop over each round, first filling in just the tally
        ut_rounds = self._get_round_data_without_tallyresults(raw_data.rounds)

        # Then, compute the tally results
        self._fill_in_tallyresults_from_tally(raw_data.rounds, ut_rounds)

        return {'config': config, 'results': ut_rounds}

    @classmethod
    def _get_round_data_without_tallyresults(cls, rounds):
        """ Generates the ut_rounds data without tallyResults """
        ut_rounds = []
        for round_i, round_data in enumerate(rounds):
            ut_round = {
                'round': round_i + 1,
                'tally': {}
            }
            candidates = round_data['candidates']
            for candidate, votes in candidates.items():
                ut_round['tally'][candidate] = votes
            ut_rounds.append(ut_round)
        return ut_rounds

    @classmethod
    def _fill_in_tallyresults_from_tally(cls, rounds, ut_rounds):
        """ Fills in tallyResults in ut_rounds """
        already_elected_set = set()
        for round_i in range(len(rounds)):
            # Get who was elected and eliminated
            elected_names = cls._get_elected_names(rounds, round_i, already_elected_set)
            eliminated_names = cls._get_eliminated_names(rounds, round_i)
            # Get how the votes change between this round and next
            vote_delta = cls.compute_vote_deltas_for_round(ut_rounds, round_i)
            # Use that to compute the tallyResults structure
            tally_results = cls.guess_at_tally_results(eliminated_names, elected_names, vote_delta)
            # Store it, completing the structure for this round
            ut_rounds[round_i]['tallyResults'] = tally_results

    @classmethod
    def _threshold_for_round(cls, rounds, round_i):
        """
        If the threshold is not provided in each round, assume it is half of the number
        of voters present in the last round. This means that if the thresold is not
        provided, we assume it is a single-winner election (unless there is an exact 50/50 tie)
        TODO: This should probably be handled by a migration function, since
        ElectionBuddy no longer outputs data without thresholds.
        """
        if rounds[round_i]['threshold']:
            return rounds[round_i]['threshold']

        last_round = rounds[-1]
        num_voters_in_last_round = sum([v for k, v in last_round['candidates'].items()])
        return num_voters_in_last_round / 2.0

    @classmethod
    def _get_eliminated_names(cls, rounds, round_i):
        if round_i == len(rounds) - 1:
            return []

        thisround_candidates = set(rounds[round_i]['candidates'])
        nextround_candidates = set(rounds[round_i + 1]['candidates'])

        # Check who is no longer around next round
        return list(thisround_candidates - nextround_candidates)

    @classmethod
    def _get_elected_names(cls, rounds, round_i, already_elected_set):
        """ :return: a list of names of candidates elected on this round """
        candidates = rounds[round_i]['candidates']
        threshold = cls._threshold_for_round(rounds, round_i)
        elected_names = [name for name in candidates if candidates[name]
                         > threshold and name not in already_elected_set]
        already_elected_set.update(elected_names)
        return elected_names

    @classmethod
    def _get_threshold(cls, csv_data):
        """
        Returns the threshold for the overall election, which doesn't make a lot of sense
        for dynamic-threshold multiwinner elections, but we mark it as just the last
        threshold in the file if the file contains thresholds.
        """
        last_round = csv_data.rounds[-1]
        if last_round['threshold']:
            return last_round['threshold']

        num_voters_in_last_round = sum([v for k, v in last_round['candidates'].items()])
        return num_voters_in_last_round / 2.0



#!/usr/bin/env python

"""
A command-line interface to some of the python functionality
"""

import argparse
import json
import sys

from enum import Enum

from rcvformats import schemas
from rcvformats.conversions.automatic import AutomaticConverter
from rcvformats.conversions.ut_without_transfers import UTWithoutTransfersConverter


class FormatEnum(Enum):
    """ Enum for validator, and perhaps eventually converter. """
    UT = 'ut'
    EB = 'eb'
    OV = 'ov'

    def __str__(self):
        return str(self.value)


def convert(input_filename, output_filename):
    """ Automatic converter from input_filename to output_filename """
    standardized_format = AutomaticConverter().convert_to_ut(input_filename)
    with open(output_filename, 'w', encoding='utf-8') as file_obj:
        json.dump(standardized_format, file_obj)


def validate(input_filename, schema):
    """ validates input_filename with schema """
    if schema == FormatEnum.UT:
        schema = schemas.universaltabulator.SchemaV0()
    elif schema == FormatEnum.EB:
        schema = schemas.electionbuddy.SchemaV0()
    elif schema == FormatEnum.OV:
        schema = schemas.opavote.SchemaV1_0()

    is_valid = schema.validate(input_filename)
    if is_valid:
        print("Schema is valid.")
    else:
        print("Schema is not valid. Errors: ", schema.last_error())


def add_transfers(input_filename, output_filename, allow_guessing):
    """ Adds tally transfers if they don't exist. Overwrites them if they do. """
    # Adding transfers, internally, is just another conversion
    converter = UTWithoutTransfersConverter(allow_guessing=allow_guessing)
    with_transfers = converter.convert_to_ut(input_filename)
    with open(output_filename, 'w', encoding='utf-8') as file_obj:
        json.dump(with_transfers, file_obj)


def _add_input_arg(parser):
    parser.add_argument(
        '-i',
        '--input',
        dest='input_filename',
        help='File to convert',
        required=True)


def _add_output_arg(parser):
    parser.add_argument(
        '-o',
        '--output',
        dest='output_filename',
        help='Where to place the JSON file on success',
        required=True)


def main():
    """ Main function: cli entrypoint, using argparse """
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(
        dest='subparser',
        help='Tooling for RCV Formats. See more at https://rcvformats.readthedocs.io/en/latest/')

    conv_parser = subparsers.add_parser(
        'convert', help='Converts from whatever format you have to the Universal Tabulator format.')
    _add_input_arg(conv_parser)
    _add_output_arg(conv_parser)

    validate_parser = subparsers.add_parser(
        'validate', help='Validates the file with one of the three accepted formats')
    _add_input_arg(validate_parser)
    validate_parser.add_argument(
        '-s',
        '--schema',
        dest='schema',
        type=FormatEnum,
        choices=list(FormatEnum),
        help='Schema to validate against: ut (Universal Tabulator/RCTab), '
             'eb (ElectionBuddy), or ov (OpaVote)',
        required=True)

    xfer_parser = subparsers.add_parser(
        'transfer', help='Adds "transfers" to the tallyResults of an otherwise-valid UI format')
    _add_input_arg(xfer_parser)
    _add_output_arg(xfer_parser)
    xfer_parser.add_argument(
        '-g',
        '--allow-guessing',
        dest='allow_guessing',
        action='store_true',
        help='During batch elimination, may we guess at the vote transfers? '
             'If not, will leave transfers blank for all batch elimination rounds.',
        required=False)

    args = parser.parse_args()
    if args.subparser is None:
        print(parser.print_help())
        sys.exit(-1)

    if args.subparser == 'convert':
        convert(args.input_filename, args.output_filename)
    if args.subparser == 'validate':
        validate(args.input_filename, args.schema)
    if args.subparser == 'transfer':
        add_transfers(args.input_filename, args.output_filename, args.allow_guessing)
